class AFile:
	"""
	Utility class since most OSes don't yet support async file read operations
	"""
	def __init__(self, filepath):
		self.filepath = filepath
		self.fd = open(filepath, 'rb')
		
	async def seek(self, pos, whence = 0):
		return self.fd.seek(pos, whence)
	
	async def read(self, n = -1):
		return self.fd.read(n)
	
	def tell(self):
		return self.fd.tell()
	
	async def close(self):
		return self.fd.close()



import io

class KERB_STORED_CREDENTIAL_NEW:
	def __init__(self):
		self.Revision = None
		self.Flags = None
		self.CredentialCount = None
		self.ServiceCredentialCount = None
		self.OldCredentialCount = None
		self.OlderCredentialCount = None
		self.DefaultSaltLength = None
		self.DefaultSaltMaximumLength = None
		self.DefaultSaltOffset = None
		self.DefaultIterationCount = None
		self.Buffer = None
	
	@staticmethod
	def from_bytes(data):
		return KERB_STORED_CREDENTIAL_NEW.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = KERB_STORED_CREDENTIAL_NEW()
		res.Revision = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.Flags = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.CredentialCount = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.ServiceCredentialCount = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.OldCredentialCount = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.OlderCredentialCount = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.DefaultSaltLength = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.DefaultSaltMaximumLength = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.DefaultSaltOffset = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.DefaultIterationCount = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.Buffer = buff.read()
		return res

	def __str__(self):
		t = ''
		for k in self.__dict__:
			t += '%s : %s\r\n' % (k, self.__dict__[k])
		return t



class KERB_KEY_DATA_NEW:
	def __init__(self):
		self._len = 24
		self.Reserved1 = None
		self.Reserved2 = None
		self.Reserved3 = None
		self.IterationCount = None
		self.KeyType = None
		self.KeyLength = None
		self.KeyOffset = None
	
	@staticmethod
	def from_bytes(data):
		return KERB_KEY_DATA_NEW.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = KERB_KEY_DATA_NEW()
		res.Reserved1 = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.Reserved2 = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.Reserved3 = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.IterationCount = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.KeyType = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.KeyLength = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.KeyOffset = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		return res

	def __str__(self):
		t = ''
		for k in self.__dict__:
			t += '%s : %s\r\n' % (k, self.__dict__[k])
		return t


import asyncio
from tqdm import tqdm
from aesedb.eesentdb import ESENT_DB
from aesedb.utils.afile import AFile
from aesedb.security.ntds import NTDS
from aesedb.utils.systemhive import SYSTEM
from aiowinreg.ahive import AIOWinRegHive
from aesedb import logger

class NTDSParserConsole:
	def __init__(self, bootkey, ntdsfile, show_progress = True, outfile = None, ext_result_q = None):
		self.bootkey = bootkey
		self.ntdsfile = ntdsfile
		self.show_progress = show_progress
		self.outfile = outfile
		self.ext_result_q = ext_result_q
		self.buffer_size = 100
		self.buffer = []

		self.outfile_handle = None
	
	async def flush_buffer(self):
		try:
			if self.outfile_handle is not None:
				res = ''
				for secret in self.buffer:
					res += str(secret)

				self.outfile_handle.write(res)

			else:
				for secret in self.buffer:			
					print(str(secret))
			
			self.buffer = []
			return True, None
		except Exception as e:
			return None, e

	async def run(self):
		try:
			hexkey = False
			if isinstance(self.bootkey, str) and len(self.bootkey) == 32:
				try:
					bootkey = bytes.fromhex(self.bootkey)
					hexkey = True
				except:
					pass
			
			if hexkey is False:
				regfile = self.bootkey
				if isinstance(self.bootkey, str):
					regfile = AFile(self.bootkey)
				reghive = AIOWinRegHive(regfile)
				syshive = SYSTEM(reghive)
				bootkey = await syshive.get_bootkey()
			
			file = self.ntdsfile
			if isinstance(self.ntdsfile, str):
				file = AFile(self.ntdsfile)
			
			db = ESENT_DB(file)
			_, err = await db.parse()
			if err is not None:
				raise err

			logger.debug('Fetching total row count')
			total, err = await db.get_rowcnt('datatable')
			if err is not None:
				raise err
			
			ntds = NTDS(db, bootkey)
			if self.show_progress is True:
				pbar     = tqdm(desc='JET table parsing ', total=total, unit='records', miniters= total//200 ,position=0)
				pbar_sec = tqdm(desc='User secrets found', unit = '', miniters=self.buffer_size//10 ,position=1)

			if self.outfile is not None:
				self.outfile_handle = open(self.outfile, 'w', newline = '')
			
			logger.debug('Dumping secrets')
			async for secret, err in ntds.dump_secrets():
				if err is not None:
					raise err

				if self.show_progress is True:
					pbar.update()
				
				if self.ext_result_q is not None:
					await self.ext_result_q.put((secret, total, None, False)) # secret, total, error, end
				
				if secret is None:
					continue
				
				if self.show_progress is True:
					pbar_sec.update()
				
				if self.ext_result_q is None:
					self.buffer.append(secret)
					if len(self.buffer) > self.buffer_size:
						_, err = await self.flush_buffer()
						if err is not None:
							raise err

			if self.ext_result_q is None:
				_, err = await self.flush_buffer()
				if err is not None:
					raise err
			
			else:
				await self.ext_result_q.put((None, total, None, True))

			return True, None
		except Exception as e:
			if self.ext_result_q is not None:
				await self.ext_result_q.put((None, None, e, True)) # secret, error, end
			return None, e

def main():
	import argparse
	import traceback
	try:
		parser = argparse.ArgumentParser(description='NTDS.dit parser')
		parser.add_argument('-v', '--verbose', action='count', default=0)
		parser.add_argument('-p', '--progress', action='store_true', help='Show progress bar')
		parser.add_argument('-o', '--out-file', help='Output file path.')
		parser.add_argument('bootkey',  help='Bootkey OR path to the SYSTEM hive file')
		parser.add_argument('ntdsfile', help='NTDS.dit file path')

		args = parser.parse_args()

		ntdscon = NTDSParserConsole(
			args.bootkey,
			args.ntdsfile,
			show_progress = args.progress,
			outfile = args.out_file,
			ext_result_q = None
		)

		_, err = asyncio.run(ntdscon.run())
		if err is not None:
			raise err
	except Exception as e:
		traceback.print_exc()

if __name__ == '__main__':
	main()
	


import io

class PEKLIST_ENC:
	def __init__(self):
		self.Header = None
		self.KeyMaterial = None
		self.EncryptedPek = None

	@staticmethod
	def from_bytes(data):
		return PEKLIST_ENC.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = PEKLIST_ENC()
		res.Header = buff.read(8)
		res.KeyMaterial= buff.read(16)
		res.EncryptedPek= buff.read()
		return res
	
	def __str__(self):
		t = ''
		for k in self.__dict__:
			t += '%s : %s\r\n' % (k, self.__dict__[k])
		return t


class PEKLIST_PLAIN:
	def __init__(self):
		self.Header = None
		self.DecryptedPek = None

	@staticmethod
	def from_bytes(data):
		return PEKLIST_PLAIN.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = PEKLIST_PLAIN()
		res.Header = buff.read(32)
		res.DecryptedPek= buff.read()
		return res

class PEK_KEY:
	def __init__(self):
		self._len = 20
		self.Header = None
		self.Padding = None
		self.Key = None

	@staticmethod
	def from_bytes(data):
		return PEK_KEY.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = PEK_KEY()
		res.Header = buff.read(1)
		res.Padding= buff.read(3)
		res.Key= buff.read(16)
		return res




from Cryptodome.Cipher import AES

import struct


# https://tools.ietf.org/html/rfc3610

def aesCCMEncrypt(plaintext, aad, key, nonce, macLen):
	blockSize = 16 # For AES...
	if macLen not in (4, 6, 8, 10, 12, 14, 16):
		raise ValueError("Parameter 'mac_len' must be even and in the range 4..16 (not %d)" % macLen)
	if not (nonce and 7 <= len(nonce) <= 13):
		raise ValueError("Length of parameter 'nonce' must be in the range 7..13 bytes")
			

	# Encryption
	q = 15 - len(nonce) 
	cipher = AES.new(key=key, mode=AES.MODE_CTR, nonce=struct.pack("B", q - 1) + nonce)
	s0 = cipher.encrypt(b'\x00'*16) # For mac
	c =  cipher.encrypt(plaintext)
	
	# Mac
	pLen = len(plaintext)
	aadLen = len(aad)
	mac = AES.new(key=key, mode=AES.MODE_CBC, iv=b'\x00'*blockSize)
	flags = (64 * (aadLen > 0) + 8 * ((macLen - 2) // 2) + (q - 1))
	b0 = struct.pack("B", flags) + nonce +  pLen.to_bytes(q, 'big')
	
	
	assocLenEncoded = b''
	if aadLen > 0:
		if aadLen < (2 ** 16 - 2 ** 8):
			encsize = 2
		elif aadLen < (2 ** 32):
			assocLenEncoded = b'\xFF\xFE'
			encsize = 4
		else:
			assocLenEncoded = b'\xFF\xFF'
			encsize = 8
		assocLenEncoded += aadLen.to_bytes(encsize, 'big')
		
	print(f"Length of block 0: {len(b0)}" )
	aadPadded =  assocLenEncoded + aad 
	aadPadded += b'\x00'*(blockSize - (len(aadPadded) % blockSize))
	ptxtPadded = plaintext + b'\x00'*(blockSize - (pLen % blockSize))
	macData = b0  + aadPadded + ptxtPadded
	print(f"Length of MAC input {len(macData)}")
	t = mac.encrypt(macData)[-16:]
	tag = bytes([a ^ b for (a,b) in zip(t,s0)])
	return (c, tag)



if __name__ == '__main__':
	import os
	EncryptionKey = b'\xff'*16
	nonce = os.urandom(11)
	hdr = b'\xFF'* 40
	msg_data = b'\xAA' * 1024

	cipher = AES.new(EncryptionKey, AES.MODE_CCM, nonce)
	cipher.update(hdr)
	enc_data = cipher.encrypt(msg_data)
	sig = cipher.digest()
	
	print(enc_data.hex())
	print(sig.hex())


	c, t = aesCCMEncrypt(msg_data, hdr, EncryptionKey, nonce, 16)

	print('other')
	print(c.hex())
	print(t.hex())

	if c == enc_data:
		print('Encryption OK!')
	else:
		print('Encryption fail')




##this file is present to force pure ciphers to be used
## reason: pyinstaller cant do dynamic imports
## usage: overwrite the "symmetric.py" file with tihs one

from aesedb.security.crypto.AES import pyCryptodomeAES
from aesedb.security.crypto.DES import pyCryptodomeDES
from aesedb.security.crypto.RC4 import pyCryptodomeRC4
from aesedb.security.crypto.AESCCM_dome import aesCCMEncrypt as ae
from aesedb.security.crypto.AESCCM_dome import aesCCMDecrypt as ad

DES = pyCryptodomeDES
AES = pyCryptodomeAES
RC4 = pyCryptodomeRC4

aesCCMEncrypt = ae
aesCCMDecrypt = ad


import importlib
import importlib.util

#
##key = name of the cipher, value=list of module names, in order of preference
preftable = {
	'DES' : ['pyCryptodome','pyCrypto','pure'], 
	'TDES': ['pyCryptodome','pyCrypto','pure'], 
	'AES' : ['pyCryptodome','cryptography','pyCrypto','pure'], 
	'RC4' : ['pyCryptodome','cryptography','pyCrypto','pure'],
}

#
available_modules = ['pure']

if importlib.util.find_spec("cryptography") is not None:
	#print('Found cryptography package!')
	available_modules.append("cryptography")

if importlib.util.find_spec("pyCrypto") is not None:
	#print('Found cryptography package!')
	available_modules.append("pyCrypto")

if importlib.util.find_spec("Cryptodome") is not None:
	#print('Found cryptography package!')
	available_modules.append("pyCryptodome")

#print(available_modules)

#https://stackoverflow.com/questions/8790003/dynamically-import-a-method-in-a-file-from-a-string
def import_from(module, name):
	module = __import__(module, fromlist=[name])
	return getattr(module, name)


def getPreferredCipher(cipherName):
	if cipherName not in preftable:
		raise Exception('Cipher %s doesnt have any preferences set!' % cipherName)
	#print('available_modules %s' % available_modules)
	possible_prefmodule = list(set(preftable[cipherName]).intersection(set(available_modules)))
	#print('possible_prefmodule %s' % possible_prefmodule)
	selected_module = None
	for moduleName in preftable[cipherName]:
		if moduleName in possible_prefmodule:
			selected_module = moduleName
			break

	if selected_module is None:
		raise Exception('Could not find any modules to load cipher %s' % cipherName)
	
	#print('Preferred module selected for cipher %s is %s' % (cipherName, selected_module))
	moduleName = 'aesedb.security.crypto.%s' % cipherName
	objectName = selected_module + cipherName
	return import_from(moduleName , objectName)

def getSpecificCipher(cipherName, moduleBaseName):
	moduleName = 'aesedb.security.crypto.%s' % cipherName
	objectName = '%s%s' % (moduleBaseName, cipherName)
	return import_from(moduleName , objectName)

def get_ccm_enc():
	if 'pyCryptodome' in available_modules:
		from aesedb.security.crypto.AESCCM_dome import aesCCMEncrypt, aesCCMDecrypt
		return (aesCCMEncrypt, aesCCMDecrypt)
	
	else:
		from aesedb.security.crypto.pure.AES.AESCCM import aesCCMEncrypt, aesCCMDecrypt
		return (aesCCMEncrypt, aesCCMDecrypt)

def get_gcm_enc():
	return None, None

#import ciphers
# TODO: fix the dynamic imports, currently only supporting pure-python ciphers for two reasons:
# 1. dynamic import messes up some scripts like pyinstaller/nuitka/py2exe
# 2. additional effort needed to support more crypto libs anyhow



from aesedb.security.crypto.DES import pureDES

DES  = getPreferredCipher('DES')
AES  = getPreferredCipher('AES')
RC4  = getPreferredCipher('RC4')

#TDES = getPreferredCipher('TDES')

aesCCMEncrypt, aesCCMDecrypt = get_ccm_enc()
#aesGCMEncrypt, aesGCMDecrypt = get_gcm_enc()


#from aesedb.security.crypto.AES import pureAES
#from aesedb.security.crypto.DES import pureDES
#from aesedb.security.crypto.DES import pureRC4

#DES = pureDES #getPreferredCipher('DES')
#AES = pureAES #getPreferredCipher('AES')
#RC4 = pureRC4




from aesedb.structures.pageheader import readheader
from aesedb.structures.rootheader import ESENT_ROOT_HEADER
from aesedb.structures.branchheader import ESENT_BRANCH_HEADER
from aesedb.structures.spacetreeheader import ESENT_SPACE_TREE_HEADER
from aesedb.structures.spacetreeentry import ESENT_SPACE_TREE_ENTRY
from aesedb.structures.indexentry import ESENT_INDEX_ENTRY
from aesedb.structures.datadefinitionheader import ESENT_DATA_DEFINITION_HEADER
from aesedb.structures.leafentry import ESENT_LEAF_ENTRY
from aesedb.structures.branchentry import ESENT_BRANCH_ENTRY
from aesedb.structures.catalogdatadefentry import ESENT_CATALOG_DATA_DEFINITION_ENTRY

from aesedb.structures.leafheader import ESENT_LEAF_HEADER

from aesedb.utils.hexdump import hexdump
import enum
from struct import unpack


class WELLKNOWNPAGENO(enum.Enum):
	DATABASE           = 1
	CATALOG            = 4
	CATALOG_BACKUP     = 24

class PAGEFLAGS(enum.IntEnum):
	ROOT         = 1
	LEAF         = 2
	PARENT       = 4
	EMPTY        = 8
	SPACE_TREE   = 0x20
	INDEX        = 0x40
	LONG_VALUE   = 0x80
	NEW_FORMAT   = 0x2000
	NEW_CHECKSUM = 0x2000


class ESENT_PAGE:
	def __init__(self, db):
		self.db = db
		self.data = None
		self.record = None
		self.tags = {} # tagno -> (flags, data)
	
	async def read(self, data):
		try:
			self.data = data
			self.record, err = await readheader(self.db.Version, self.db.FileFormatRevision, self.db.PageSize, self.data)
			if err is not None:
				raise err
			return True, None
		except Exception as e:
			return None, e

	def printFlags(self):
		print(self.record.PageFlags)

	def dump(self):
		baseOffset = len(self.record)
		print(self.record)
		tags = self.data[-4*self.record.FirstAvailablePageTag:]

		print("FLAGS: ")
		self.printFlags()

		print()

		for i in range(self.record.FirstAvailablePageTag):
			tag = tags[-4:]
			if self.db.Version == 0x620 and self.db.FileFormatRevision > 11 and self.db.PageSize > 8192:
				valueSize = unpack('<H', tag[:2])[0] & 0x7fff
				valueOffset = unpack('<H',tag[2:])[0] & 0x7fff
				hexdump((self.data[baseOffset+valueOffset:][:6]))
				pageFlags = ord(self.data[baseOffset+valueOffset:][1]) >> 5
				#print "TAG FLAG: 0x%x " % (unpack('<L', self.data[baseOffset+valueOffset:][:4]) ) >> 5
				#print "TAG FLAG: 0x " , ord(self.data[baseOffset+valueOffset:][0])
			else:
				valueSize = unpack('<H', tag[:2])[0] & 0x1fff
				pageFlags = (unpack('<H', tag[2:])[0] & 0xe000) >> 13
				valueOffset = unpack('<H',tag[2:])[0] & 0x1fff
				
			print("TAG %-8d offset:0x%-6x flags:0x%-4x valueSize:0x%x" % (i,valueOffset,pageFlags,valueSize))
			#hexdump(self.getTag(i)[1])
			tags = tags[:-4]

		if self.record.PageFlags & PAGEFLAGS.ROOT > 0:
			rootHeader = ESENT_ROOT_HEADER(self.getTag(0)[1])
			print(rootHeader)
		elif self.record.PageFlags & PAGEFLAGS.LEAF == 0:
			# Branch Header
			flags, data = self.getTag(0)
			branchHeader = ESENT_BRANCH_HEADER(data)
			print(branchHeader)
		else:
			# Leaf Header
			flags, data = self.getTag(0)
			if self.record.PageFlags & PAGEFLAGS.SPACE_TREE > 0:
				# Space Tree
				spaceTreeHeader = ESENT_SPACE_TREE_HEADER(data)
				print(spaceTreeHeader)
			else:
				leafHeader = ESENT_LEAF_HEADER(data)
				print(leafHeader)

		# Print the leaf/branch tags
		for tagNum in range(1,self.record.FirstAvailablePageTag):
			flags, data = self.getTag(tagNum)
			if self.record.PageFlags & PAGEFLAGS.LEAF == 0:
				# Branch page
				branchEntry = ESENT_BRANCH_ENTRY(flags, data)
				print(branchEntry)
			elif self.record.PageFlags & PAGEFLAGS.LEAF > 0:
				# Leaf page
				if self.record.PageFlags & PAGEFLAGS.SPACE_TREE > 0:
					# Space Tree
					spaceTreeEntry = ESENT_SPACE_TREE_ENTRY(data)
					#print(spaceTreeEntry)

				elif self.record.PageFlags & PAGEFLAGS.INDEX > 0:
					# Index Entry
					indexEntry = ESENT_INDEX_ENTRY(data)
					#print(indexEntry)
				elif self.record.PageFlags & PAGEFLAGS.LONG_VALUE > 0:
					# Long Page Value
					raise Exception('Long value still not supported')
				else:
					# Table Value
					leafEntry = ESENT_LEAF_ENTRY(flags, data)
					dataDefinitionHeader = ESENT_DATA_DEFINITION_HEADER.from_bytes(leafEntry.EntryData)
					print(dataDefinitionHeader)
					catalogEntry = ESENT_CATALOG_DATA_DEFINITION_ENTRY.from_bytes(leafEntry.EntryData[len(dataDefinitionHeader):])
					print(catalogEntry)
					hexdump(leafEntry.EntryData)

	def get_tag(self, tagno):
		if self.record.FirstAvailablePageTag < tagno:
			raise Exception('Trying to grab an unknown tag 0x%x' % tagno)

		if tagno in self.tags:
			return self.tags[tagno]

		tags = self.data[-4*self.record.FirstAvailablePageTag:]
		baseOffset = self.record._len
		
		#input(tags)
		tag = tags[(-4*tagno)-4:-4*tagno]
		#input(calc_tag)

		#tag = tagno*4
		
		#### original
		#tags = self.data[-4*self.record.FirstAvailablePageTag:]
		#for i in range(tagno):
		#	tags = tags[:-4]
		#tag = tags[-4:]

		#input(tag)



		

		if self.db.Version == 0x620 and self.db.FileFormatRevision >= 17 and self.db.PageSize > 8192:
			valueSize = unpack('<H', tag[:2])[0] & 0x7fff
			valueOffset = unpack('<H',tag[2:])[0] & 0x7fff
			tmpData = list(self.data[baseOffset+valueOffset:][:valueSize])
			pageFlags = ord(tmpData[1]) >> 5
			tmpData[1] = chr(ord(tmpData[1:2]) & 0x1f)
			tagData = "".join(tmpData)
		else:
			valueSize = unpack('<H', tag[:2])[0] & 0x1fff
			pageFlags = (unpack('<H', tag[2:])[0] & 0xe000) >> 13
			valueOffset = unpack('<H',tag[2:])[0] & 0x1fff
			tagData = self.data[baseOffset+valueOffset:][:valueSize]
		
		self.tags[tagno] = (pageFlags, tagData)
		#return pageFlags, self.data[baseOffset+valueOffset:][:valueSize]
		return pageFlags, tagData


from Cryptodome.Cipher import AES

def aesCCMEncrypt(plaintext, aad, key, nonce, macLen = 16):
	cipher = AES.new(key, AES.MODE_CCM,  nonce)
	cipher.update(aad)
	ciphertext = cipher.encrypt(plaintext)
	mac = cipher.digest()
	return ciphertext, mac

def aesCCMDecrypt(ciphertext, aad, key, nonce, macValue):
	cipher = AES.new(key, AES.MODE_CCM,  nonce)
	cipher.update(aad)
	plaintext = cipher.encrypt(ciphertext)
	return plaintext




import io

class ESENT_DATA_DEFINITION_HEADER:
	def __init__(self):
		self._len = 4
		self.LastFixedSize = None
		self.LastVariableDataType = None
		self.VariableSizeOffset = None

	@staticmethod
	def from_bytes(data):
		return ESENT_DATA_DEFINITION_HEADER.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = ESENT_DATA_DEFINITION_HEADER()
		res.LastFixedSize = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		res.LastVariableDataType = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		res.VariableSizeOffset = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		return res



import io

class ESENT_INDEX_ENTRY:
	def __init__(self):
		self.RecordPageKey = None

	@staticmethod
	def from_bytes(data):
		return ESENT_INDEX_ENTRY.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = ESENT_INDEX_ENTRY()
		res.RecordPageKey = buff.read()
		return res


from Cryptodome.Cipher import AES

def aesGCMEncrypt(plaintext, aad, key, nonce, macLen = 16):
	cipher = AES.new(key, AES.MODE_GCM,  nonce=nonce)
	cipher.update(aad)
	return cipher.encrypt_and_digest(plaintext)

def aesGCMDecrypt(ciphertext, aad, key, nonce, macValue):
	cipher = AES.new(key, AES.MODE_GCM,  nonce=nonce)
	cipher.update(aad)
	plaintext = cipher.decrypt(ciphertext)
	return plaintext



import io
from aesedb.structures.jetsignature import ESENT_JET_SIGNATURE

class ESENT_DB_HEADER:
	def __init__(self):
		self.CheckSum = None
		self.Signature = None
		self.Version= None
		self.FileType= None
		self.DBTime= None
		self.DBSignature= None
		self.DBState= None
		self.ConsistentPosition= None
		self.ConsistentTime= None
		self.AttachTime= None
		self.AttachPosition= None
		self.DetachTime= None
		self.DetachPosition= None
		self.LogSignature= None
		self.Unknown= None
		self.PreviousBackup= None
		self.PreviousIncBackup= None
		self.CurrentFullBackup= None
		self.ShadowingDisables= None
		self.LastObjectID= None
		self.WindowsMajorVersion= None
		self.WindowsMinorVersion= None
		self.WindowsBuildNumber= None
		self.WindowsServicePackNumber= None
		self.FileFormatRevision= None
		self.PageSize= None
		self.RepairCount= None
		self.RepairTime= None
		self.Unknown2= None
		self.ScrubTime= None
		self.RequiredLog= None
		self.UpgradeExchangeFormat= None
		self.UpgradeFreePages= None
		self.UpgradeSpaceMapPages= None
		self.CurrentShadowBackup= None
		self.CreationFileFormatVersion= None
		self.CreationFileFormatRevision= None
		self.Unknown3= None
		self.OldRepairCount= None
		self.ECCCount= None
		self.LastECCTime= None
		self.OldECCFixSuccessCount= None
		self.ECCFixErrorCount= None
		self.LastECCFixErrorTime= None
		self.OldECCFixErrorCount= None
		self.BadCheckSumErrorCount= None
		self.LastBadCheckSumTime= None
		self.OldCheckSumErrorCount= None
		self.CommittedLog= None
		self.PreviousShadowCopy= None
		self.PreviousDifferentialBackup= None
		self.Unknown4= None
		self.NLSMajorVersion= None
		self.NLSMinorVersion= None
		self.Unknown5= None
		self.UnknownFlags= None

	
	@staticmethod
	def from_bytes(data):
		return ESENT_DB_HEADER.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = ESENT_DB_HEADER()
		res.CheckSum = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.Signature = buff.read(4)
		res.Version = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.FileType = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.DBTime = int.from_bytes(buff.read(8), byteorder='little', signed=False)
		res.DBSignature = ESENT_JET_SIGNATURE.from_buffer(buff)
		res.DBState = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.ConsistentPosition= int.from_bytes(buff.read(8), byteorder='little', signed=False)
		res.ConsistentTime= int.from_bytes(buff.read(8), byteorder='little', signed=False)
		res.AttachTime= int.from_bytes(buff.read(8), byteorder='little', signed=False)
		res.AttachPosition= int.from_bytes(buff.read(8), byteorder='little', signed=False)
		res.DetachTime= int.from_bytes(buff.read(8), byteorder='little', signed=False)
		res.DetachPosition= int.from_bytes(buff.read(8), byteorder='little', signed=False)
		res.LogSignature= ESENT_JET_SIGNATURE.from_buffer(buff)
		res.Unknown= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.PreviousBackup= buff.read(24)
		res.PreviousIncBackup= buff.read(24)
		res.CurrentFullBackup= buff.read(24)
		res.ShadowingDisables= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.LastObjectID = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.WindowsMajorVersion= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.WindowsMinorVersion= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.WindowsBuildNumber= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.WindowsServicePackNumber= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.FileFormatRevision= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.PageSize = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.RepairCount= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.RepairTime= int.from_bytes(buff.read(8), byteorder='little', signed=False)
		res.Unknown2= buff.read(28)
		res.ScrubTime= int.from_bytes(buff.read(8), byteorder='little', signed=False)
		res.RequiredLog= int.from_bytes(buff.read(8), byteorder='little', signed=False)
		res.UpgradeExchangeFormat= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.UpgradeFreePages= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.UpgradeSpaceMapPages= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.CurrentShadowBackup= buff.read(24)
		res.CreationFileFormatVersion= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.CreationFileFormatRevision= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.Unknown3= buff.read(16)
		res.OldRepairCount= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.ECCCount= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.LastECCTime= int.from_bytes(buff.read(8), byteorder='little', signed=False)
		res.OldECCFixSuccessCount= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.ECCFixErrorCount= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.LastECCFixErrorTime= int.from_bytes(buff.read(8), byteorder='little', signed=False)
		res.OldECCFixErrorCount= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.BadCheckSumErrorCount= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.LastBadCheckSumTime= int.from_bytes(buff.read(8), byteorder='little', signed=False)
		res.OldCheckSumErrorCount= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.CommittedLog= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.PreviousShadowCopy= buff.read(24)
		res.PreviousDifferentialBackup= buff.read(24)
		res.Unknown4= buff.read(40)
		res.NLSMajorVersion= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.NLSMinorVersion= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.Unknown5= buff.read(148)
		res.UnknownFlags= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		return res
	
	def __str__(self) -> str:
		t = ''
		for k in self.__dict__:
			t += '%s : %s\r\n' % (k, self.__dict__[k])
		return t


import io

class ESENT_ROOT_HEADER:
	def __init__(self):
		self.InitialNumberOfPages = None
		self.ParentFatherDataPage = None
		self.ExtentSpace= None
		self.SpaceTreePageNumber= None
	
	@staticmethod
	def from_bytes(data):
		return ESENT_ROOT_HEADER.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = ESENT_ROOT_HEADER()
		res.InitialNumberOfPages = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.ParentFatherDataPage = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.ExtentSpace= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.SpaceTreePageNumber= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		return res




##this file is present to force pure ciphers to be used
## reason: pyinstaller cant do dynamic imports
## usage: overwrite the "symmetric.py" file with tihs one

from aesedb.security.crypto.AES import pureAES
from aesedb.security.crypto.DES import pureDES
from aesedb.security.crypto.RC4 import pureRC4
from aesedb.security.crypto.pure.AES.AESCCM import aesCCMEncrypt as ae
from aesedb.security.crypto.pure.AES.AESCCM import aesCCMDecrypt as ad

DES = pureDES
AES = pureAES
RC4 = pureRC4

aesCCMEncrypt = ae
aesCCMDecrypt = ad




import io
import enum

class CATALOG_TYPE(enum.Enum):
	TABLE        = 1
	COLUMN       = 2
	INDEX        = 3
	LONG_VALUE   = 4
	CALLBACK     = 5

class ESENT_CATALOG_DATA_DEFINITION_ENTRY:
	@staticmethod
	def from_bytes(data):
		return ESENT_CATALOG_DATA_DEFINITION_ENTRY.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		pos = buff.tell()
		buff.seek(4)
		dtype = CATALOG_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed=False))
		buff.seek(pos)
		if dtype == CATALOG_TYPE.COLUMN:
			return ESENT_CATALOG_DATA_DEFINITION_ENTRY_COL.from_buffer(buff)
		elif dtype == CATALOG_TYPE.INDEX:
			return ESENT_CATALOG_DATA_DEFINITION_ENTRY_INDEX.from_buffer(buff)
		elif dtype == CATALOG_TYPE.TABLE:
			return ESENT_CATALOG_DATA_DEFINITION_ENTRY_TABLE.from_buffer(buff)
		elif dtype == CATALOG_TYPE.LONG_VALUE:
			return ESENT_CATALOG_DATA_DEFINITION_ENTRY_LV.from_buffer(buff)
		else:
			raise Exception('Unsupported type %s' % dtype)

class ESENT_CATALOG_DATA_DEFINITION_ENTRY_COL:
	def __init__(self):
		self.FatherDataPageID = None
		self.Type = None
		self.Identifier = None
		self.ColumnType= None
		self.SpaceUsage= None
		self.ColumnFlags= None
		self.CodePage= None
		self.Trailing = None
	
	@staticmethod
	def from_bytes(data):
		return ESENT_CATALOG_DATA_DEFINITION_ENTRY_COL.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = ESENT_CATALOG_DATA_DEFINITION_ENTRY_COL()
		res.FatherDataPageID = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.Type = CATALOG_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed=False))
		res.Identifier = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.ColumnType= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.SpaceUsage= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.ColumnFlags= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.CodePage= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.Trailing = buff.read()
		return res

class ESENT_CATALOG_DATA_DEFINITION_ENTRY_OTHER:
	def __init__(self):
		self.FatherDataPageID = None
		self.Type = None
		self.Identifier = None
		self.FatherDataPageNumber = None
		self.Trailing = None

	
	@staticmethod
	def from_bytes(data):
		return ESENT_CATALOG_DATA_DEFINITION_ENTRY_OTHER.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = ESENT_CATALOG_DATA_DEFINITION_ENTRY_OTHER()
		res.FatherDataPageID = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.Type = CATALOG_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed=False))
		res.Identifier = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.FatherDataPageNumber= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.Trailing = buff.read()
		return res

class ESENT_CATALOG_DATA_DEFINITION_ENTRY_TABLE:
	def __init__(self):
		self.FatherDataPageID = None
		self.Type = None
		self.Identifier = None
		self.FatherDataPageNumber = None
		self.SpaceUsage = None
		self.Trailing = None

	
	@staticmethod
	def from_bytes(data):
		return ESENT_CATALOG_DATA_DEFINITION_ENTRY_TABLE.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = ESENT_CATALOG_DATA_DEFINITION_ENTRY_TABLE()
		res.FatherDataPageID = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.Type = CATALOG_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed=False))
		res.Identifier = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.FatherDataPageNumber = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.SpaceUsage= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.Trailing = buff.read()
		return res

class ESENT_CATALOG_DATA_DEFINITION_ENTRY_INDEX:
	def __init__(self):
		self.FatherDataPageID = None
		self.Type = None
		self.Identifier = None
		self.FatherDataPageNumber = None
		self.SpaceUsage = None
		self.IndexFlags = None
		self.Locale = None
		self.Trailing = None

	
	@staticmethod
	def from_bytes(data):
		return ESENT_CATALOG_DATA_DEFINITION_ENTRY_INDEX.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = ESENT_CATALOG_DATA_DEFINITION_ENTRY_INDEX()
		res.FatherDataPageID = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.Type = CATALOG_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed=False))
		res.Identifier = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.FatherDataPageNumber = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.SpaceUsage= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.IndexFlags= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.Locale= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.Trailing = buff.read()
		return res

class ESENT_CATALOG_DATA_DEFINITION_ENTRY_LV:
	def __init__(self):
		self.FatherDataPageID = None
		self.Type = None
		self.Identifier = None
		self.FatherDataPageNumber = None
		self.SpaceUsage = None
		self.Trailing = None

	@staticmethod
	def from_bytes(data):
		return ESENT_CATALOG_DATA_DEFINITION_ENTRY_LV.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = ESENT_CATALOG_DATA_DEFINITION_ENTRY_LV()
		res.FatherDataPageID = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.Type = CATALOG_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed=False))
		res.Identifier = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.FatherDataPageNumber = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.SpaceUsage= int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.Trailing = buff.read()
		return res
	


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2019 James Seo <james@equiv.tech> (github.com/kangtastic).
#
# This file is released under the WTFPL, version 2 (wtfpl.net).
#
# md4.py: An implementation of the MD4 hash algorithm in pure Python 3.
#
# Description: Zounds! Yet another rendition of pseudocode from RFC1320!
#              Bonus points for the algorithm literally being from 1992.
#
# Usage: Why would anybody use this? This is self-rolled crypto, and
#        self-rolled *obsolete* crypto at that. DO NOT USE if you need
#        something "performant" or "secure". :P
#
#        Anyway, from the command line:
#
#           $ ./md4.py [messages]
#
#        where [messages] are some strings to be hashed.
#
#        In Python, use similarly to hashlib (not that it even has MD4):
#
#           from .md4 import MD4
#
#           digest = MD4("BEES").hexdigest()
#
#           print(digest)  # "501af1ef4b68495b5b7e37b15b4cda68"
#
#
# Sample console output:
#
#   Testing the MD4 class.
#
#   Message:  b''
#   Expected: 31d6cfe0d16ae931b73c59d7e0c089c0
#   Actual:   31d6cfe0d16ae931b73c59d7e0c089c0
#
#   Message:  b'The quick brown fox jumps over the lazy dog'
#   Expected: 1bee69a46ba811185c194762abaeae90
#   Actual:   1bee69a46ba811185c194762abaeae90
#
#   Message:  b'BEES'
#   Expected: 501af1ef4b68495b5b7e37b15b4cda68
#   Actual:   501af1ef4b68495b5b7e37b15b4cda68
#
import struct


class MD4:
	"""An implementation of the MD4 hash algorithm."""

	width = 32
	mask = 0xFFFFFFFF

	# Unlike, say, SHA-1, MD4 uses little-endian. Fascinating!
	h = [0x67452301, 0xEFCDAB89, 0x98BADCFE, 0x10325476]

	def __init__(self, msg=None):
		""":param ByteString msg: The message to be hashed."""
		if msg is None:
			msg = b""

		self.msg = msg

		# Pre-processing: Total length is a multiple of 512 bits.
		ml = len(msg) * 8
		msg += b"\x80"
		msg += b"\x00" * (-(len(msg) + 8) % 64)
		msg += struct.pack("<Q", ml)

		# Process the message in successive 512-bit chunks.
		self._process([msg[i : i + 64] for i in range(0, len(msg), 64)])

	def __repr__(self):
		if self.msg:
			return f"{self.__class__.__name__}({self.msg:s})"
		return f"{self.__class__.__name__}()"

	def __str__(self):
		return self.hexdigest()

	def __eq__(self, other):
		return self.h == other.h

	def bytes(self):
		""":return: The final hash value as a `bytes` object."""
		return struct.pack("<4L", *self.h)

	def hexbytes(self):
		""":return: The final hash value as hexbytes."""
		return self.hexdigest().encode

	def hexdigest(self):
		""":return: The final hash value as a hexstring."""
		return "".join(f"{value:02x}" for value in self.bytes())
	
	def digest(self):
		return self.bytes()

	def _process(self, chunks):
		for chunk in chunks:
			X, h = list(struct.unpack("<16I", chunk)), self.h.copy()

			# Round 1.
			Xi = [3, 7, 11, 19]
			for n in range(16):
				i, j, k, l = map(lambda x: x % 4, range(-n, -n + 4))
				K, S = n, Xi[n % 4]
				hn = h[i] + MD4.F(h[j], h[k], h[l]) + X[K]
				h[i] = MD4.lrot(hn & MD4.mask, S)

			# Round 2.
			Xi = [3, 5, 9, 13]
			for n in range(16):
				i, j, k, l = map(lambda x: x % 4, range(-n, -n + 4))
				K, S = n % 4 * 4 + n // 4, Xi[n % 4]
				hn = h[i] + MD4.G(h[j], h[k], h[l]) + X[K] + 0x5A827999
				h[i] = MD4.lrot(hn & MD4.mask, S)

			# Round 3.
			Xi = [3, 9, 11, 15]
			Ki = [0, 8, 4, 12, 2, 10, 6, 14, 1, 9, 5, 13, 3, 11, 7, 15]
			for n in range(16):
				i, j, k, l = map(lambda x: x % 4, range(-n, -n + 4))
				K, S = Ki[n], Xi[n % 4]
				hn = h[i] + MD4.H(h[j], h[k], h[l]) + X[K] + 0x6ED9EBA1
				h[i] = MD4.lrot(hn & MD4.mask, S)

			self.h = [((v + n) & MD4.mask) for v, n in zip(self.h, h)]

	@staticmethod
	def F(x, y, z):
		return (x & y) | (~x & z)

	@staticmethod
	def G(x, y, z):
		return (x & y) | (x & z) | (y & z)

	@staticmethod
	def H(x, y, z):
		return x ^ y ^ z

	@staticmethod
	def lrot(value, n):
		lbits, rbits = (value << n) & MD4.mask, value >> (MD4.width - n)
		return lbits | rbits


def main():
	# Import is intentionally delayed.
	import sys

	if len(sys.argv) > 1:
		messages = [msg.encode() for msg in sys.argv[1:]]
		for message in messages:
			print(MD4(message).hexdigest())
	else:
		messages = [b"", b"The quick brown fox jumps over the lazy dog", b"BEES"]
		known_hashes = [
			"31d6cfe0d16ae931b73c59d7e0c089c0",
			"1bee69a46ba811185c194762abaeae90",
			"501af1ef4b68495b5b7e37b15b4cda68",
		]

		print("Testing the MD4 class.")
		print()

		for message, expected in zip(messages, known_hashes):
			print("Message: ", message)
			print("Expected:", expected)
			print("Actual:  ", MD4(message).hexdigest())
			print()


if __name__ == "__main__":
	try:
		main()
	except KeyboardInterrupt:
		pass




import asyncio
from collections import OrderedDict
from aesedb.security.constants import NAME_TO_INTERNAL, ACCOUNT_TYPES, KERBEROS_TYPE
from aesedb.security.structures.pek import PEKLIST_ENC, PEK_KEY, PEKLIST_PLAIN
from aesedb.security.structures.crypted import CRYPTED_HASH, CRYPTED_HASHW16, CRYPTED_HISTORY, CRYPTED_BLOB
from aesedb.security.structures.userprops import USER_PROPERTIES
from aesedb.security.structures.kerbsecret import KERB_STORED_CREDENTIAL_NEW, KERB_KEY_DATA_NEW
from aesedb.security.crypto.symmetric import RC4, AES, DES
from aesedb.security.crypto.hashing import md5
from aesedb.security.crypto.BASE import cipherMODE
from aesedb.security.crypto.DES import deriveKey
from aesedb.security.structures.sid import SAMR_RPC_SID
from aesedb.security.common.usersecret import UserSecrets
from aesedb import logger
import datetime

from struct import unpack

NTDS_SUPPORTED_SECRET_TYPES = {
	NAME_TO_INTERNAL['objectSid'] : 1,
	NAME_TO_INTERNAL['dBCSPwd'] : 1,
	NAME_TO_INTERNAL['name'] : 1,
	NAME_TO_INTERNAL['sAMAccountType'] : 1,
	NAME_TO_INTERNAL['unicodePwd'] : 1,
	NAME_TO_INTERNAL['sAMAccountName'] : 1,
	NAME_TO_INTERNAL['userPrincipalName'] : 1,
	NAME_TO_INTERNAL['ntPwdHistory'] : 1,
	NAME_TO_INTERNAL['lmPwdHistory'] : 1,
	NAME_TO_INTERNAL['pwdLastSet'] : 1,
	NAME_TO_INTERNAL['userAccountControl'] : 1,
	NAME_TO_INTERNAL['supplementalCredentials'] : 1,
}


class NTDS:
	def __init__(self, db, bootkey):
		self.db = db
		self.bootkey = bootkey

		self.__PEK = []
		self.__cryptoCommon = None #CryptoCommon()
		self.__kerberosKeys = OrderedDict()
		self.__clearTextPwds = OrderedDict()
	
	async def open_db_table(self):
		try:
			logger.debug('Opening datatable...')
			cursor, err = await self.db.open_table('datatable')
			if err is not None:
				raise err
			return cursor, None
		except Exception as e:
			return None, e

	async def get_pek(self):
		try:
			logger.debug('Fetching PEK...')
			db_cursor, err = await self.open_db_table()
			if err is not None:
				raise err
			
			pklistname = NAME_TO_INTERNAL['pekList']
			peklist_enc = None
			while True:
				record, err = await self.db.get_next_row(db_cursor, {pklistname:1})
				if err is not None:
					raise err
				
				if record is None:
					raise Exception('PEK not found!')
				
				if pklistname in record and record[pklistname] is None:
					continue
				
				if pklistname in record:
					peklist_enc = PEKLIST_ENC.from_bytes(record[pklistname])
					break
			
			if peklist_enc is None:
				raise Exception('PEK not found!')

			#### decrypting peklist_enc
			if peklist_enc.Header[:4] == b'\x02\x00\x00\x00':
				# Up to Windows 2012 R2 looks like header starts this way
				
				ctx = md5(self.bootkey)
				for _ in range(1000):
					ctx.update(peklist_enc.KeyMaterial)
				tmpKey = ctx.digest()
				rc4 = RC4(tmpKey)
				peklist = PEKLIST_PLAIN.from_bytes(rc4.encrypt(peklist_enc.EncryptedPek))
				PEKLen = PEK_KEY()._len
				for i in range(len( peklist.DecryptedPek ) // PEKLen ):
					cursor = i * PEKLen
					pek = PEK_KEY.from_bytes(peklist.DecryptedPek[cursor:cursor+PEKLen])
					#print("PEK # %d found and decrypted: %s" % (i, pek['Key'].hex()))
					self.__PEK.append(pek.Key)

			elif peklist_enc.Header[:4] == b'\x03\x00\x00\x00':
				# Windows 2016 TP4 header starts this way
				# Encrypted PEK Key seems to be different, but actually similar to decrypting LSA Secrets.
				# using AES:
				# Key: the bootKey
				# CipherText: PEKLIST_ENC['EncryptedPek']
				# IV: PEKLIST_ENC['KeyMaterial']
				ctx = AES(self.bootkey, cipherMODE.CBC, IV=peklist_enc.KeyMaterial)
				dec_pek = ctx.decrypt(peklist_enc.EncryptedPek)
				peklist = PEKLIST_PLAIN.from_bytes(dec_pek)

				# PEK list entries take the form:
				#   index (4 byte LE int), PEK (16 byte key)
				# the entries are in ascending order, and the list is terminated
				# by an entry with a non-sequential index (08080808 observed)
				pos, cur_index = 0, 0
				while True:
					pek_entry = peklist.DecryptedPek[pos:pos+20]
					if len(pek_entry) < 20: break # if list truncated, should not happen
					index, pek = unpack('<L16s', pek_entry)
					if index != cur_index: break # break on non-sequential index
					self.__PEK.append(pek)
					#print("PEK # %d found and decrypted: %s" % ( index, pek.hex()))
					cur_index += 1
					pos += 20

			logger.debug('PEK found and decrypted!')
			return True, None
		
		except Exception as e:
			return None, e
	
	def __removeDESLayer(self, cryptedHash, rid):
		Key1,Key2 = deriveKey(int(rid))
		Crypt1 = DES(Key1, cipherMODE.ECB)
		Crypt2 = DES(Key2, cipherMODE.ECB)
		decryptedHash = Crypt1.decrypt(cryptedHash[:8]) + Crypt2.decrypt(cryptedHash[8:])
		return decryptedHash
	
	def __removeRC4Layer(self, cryptedHash):
		# PEK index can be found on header of each ciphered blob (pos 8-10)
		pek_ixd = cryptedHash.Header[4]
		ctx = md5(self.__PEK[pek_ixd])
		ctx.update(cryptedHash.KeyMaterial)
		tmpKey = ctx.digest()
		rc4 = RC4(tmpKey)
		plainText = rc4.encrypt(cryptedHash.EncryptedHash)
		return plainText

	@staticmethod
	def __fileTimeToDateTime(t):
		t -= 116444736000000000
		t //= 10000000
		if t < 0:
			return 'never'
		else:
			dt = datetime.datetime.fromtimestamp(t)
			return dt.strftime("%Y-%m-%d %H:%M")
	
	def __decrypt_hash(self, secret: UserSecrets, record, db_cursor, with_history = True):
		try:
			try:
				sid = SAMR_RPC_SID.from_bytes(record[NAME_TO_INTERNAL['objectSid']])
				rid = str(sid).split('-')[-1]
			except Exception as e:
				print('SID parsing failed!')
				print('Data inpuit: %s' % record[NAME_TO_INTERNAL['objectSid']].hex())
				raise e
			secret.object_sid = sid

			if record[NAME_TO_INTERNAL['dBCSPwd']] is not None:
				encryptedLMHash = CRYPTED_HASH.from_bytes(record[self.NAME_TO_INTERNAL['dBCSPwd']])
				if encryptedLMHash.Header[:4] == b'\x13\x00\x00\x00':
					# Win2016 TP4 decryption is different
					encryptedLMHash = CRYPTED_HASHW16(record[self.NAME_TO_INTERNAL['dBCSPwd']])
					pekIndex = encryptedLMHash.Header[4]
					ctx = AES(self.__PEK[pekIndex], cipherMODE.CBC, IV=encryptedLMHash.KeyMaterial)
					dec_hash_temp = ctx.decrypt(encryptedLMHash.EncryptedHash[:16])
				else:
					dec_hash_temp = self.__removeRC4Layer(encryptedLMHash)
				secret.lm_hash = self.__removeDESLayer(dec_hash_temp, rid)
			else:
				secret.lm_hash = bytes.fromhex('aad3b435b51404eeaad3b435b51404ee')

			if record[NAME_TO_INTERNAL['unicodePwd']] is not None:
				encryptedNTHash = CRYPTED_HASH.from_bytes(record[NAME_TO_INTERNAL['unicodePwd']])
				if encryptedNTHash.Header[:4] == b'\x13\x00\x00\x00':
					# Win2016 TP4 decryption is different
					encryptedNTHash = CRYPTED_HASHW16.from_bytes(record[NAME_TO_INTERNAL['unicodePwd']])
					pekIndex = encryptedNTHash.Header[4]
					ctx = AES(self.__PEK[pekIndex], cipherMODE.CBC, IV=encryptedNTHash.KeyMaterial)
					dec_hash_temp = ctx.decrypt(encryptedNTHash.EncryptedHash[:16])
				else:
					dec_hash_temp = self.__removeRC4Layer(encryptedNTHash)
				secret.nt_hash = self.__removeDESLayer(dec_hash_temp, rid)
			else:
				secret.nt_hash = bytes.fromhex('31d6cfe0d16ae931b73c59d7e0c089c0')

			if record[NAME_TO_INTERNAL['userPrincipalName']] is not None:
				secret.domain = record[NAME_TO_INTERNAL['userPrincipalName']].split('@')[-1]
				secret.username = '%s' % record[NAME_TO_INTERNAL['sAMAccountName']]
			else:
				secret.username = '%s' % record[NAME_TO_INTERNAL['sAMAccountName']]


			if record[NAME_TO_INTERNAL['userAccountControl']] is not None:
				secret.user_account_control = record[NAME_TO_INTERNAL['userAccountControl']]

			if record[NAME_TO_INTERNAL['pwdLastSet']] is not None:
				secret.pwd_last_set = self.__fileTimeToDateTime(record[NAME_TO_INTERNAL['pwdLastSet']])

			if with_history is False:
				return None, None

			if record[NAME_TO_INTERNAL['lmPwdHistory']] is not None:
				encryptedLMHistory = CRYPTED_HISTORY.from_bytes(record[NAME_TO_INTERNAL['lmPwdHistory']])
				tmpLMHistory = self.__removeRC4Layer(encryptedLMHistory)
				for i in range(0, len(tmpLMHistory) // 16):
					LMHash = self.__removeDESLayer(tmpLMHistory[i * 16:(i + 1) * 16], rid)
					secret.lm_history.append(LMHash)

			if record[NAME_TO_INTERNAL['ntPwdHistory']] is not None:
				encryptedNTHistory = CRYPTED_HISTORY.from_bytes(record[NAME_TO_INTERNAL['ntPwdHistory']])

				if encryptedNTHistory.Header[:4] == b'\x13\x00\x00\x00':
					# Win2016 TP4 decryption is different
					encryptedNTHistory = CRYPTED_HASHW16.from_bytes(record[NAME_TO_INTERNAL['ntPwdHistory']])
					pekIndex = encryptedNTHistory.Header[4]
					ctx = AES(self.__PEK[pekIndex], cipherMODE.CBC, IV=encryptedNTHistory.KeyMaterial)
					dec_hash_temp = ctx.decrypt(encryptedNTHistory.EncryptedHash)
				else:
					dec_hash_temp = self.__removeRC4Layer(encryptedNTHistory)

				for i in range(0, len(dec_hash_temp) // 16):
					NTHash = self.__removeDESLayer(dec_hash_temp[i * 16:(i + 1) * 16], rid)
					secret.nt_history.append(NTHash)

			return True, None
		except Exception as e:
			return None, e

	def __decryptSupplementalInfo(self, secret: UserSecrets, record):
		try:
			haveInfo = False
			if record[NAME_TO_INTERNAL['supplementalCredentials']] is not None:
				if len(record[NAME_TO_INTERNAL['supplementalCredentials']]) > 24:
					cipherText = CRYPTED_BLOB.from_bytes(record[NAME_TO_INTERNAL['supplementalCredentials']])
					if cipherText.Header[:4] == b'\x13\x00\x00\x00':
						# Win2016 TP4 decryption is different
						pekIndex = cipherText.Header[4]
						ctx = AES(self.__PEK[pekIndex], cipherMODE.CBC, IV=cipherText.KeyMaterial)
						plainText = ctx.decrypt(cipherText.EncryptedHash[4:])
						haveInfo = True
					else:
						plainText = self.__removeRC4Layer(cipherText)
						haveInfo = True
			
			
			if haveInfo is False:
				return True, None
			

			### woohooo we have info!
			try:
				userProperties = USER_PROPERTIES.from_bytes(plainText)
			except Exception as e:
				raise e
				# On some old w2k3 there might be user properties that don't
				# match [MS-SAMR] structure, discarding them
				return True, None

			for userprop in userProperties.UserProperties:
				if userprop.PropertyName == 'Primary:Kerberos-Newer-Keys':
					propertyValueBuffer = bytes.fromhex(userprop.PropertyValue.decode())
					kerbStoredCredentialNew = KERB_STORED_CREDENTIAL_NEW.from_bytes(propertyValueBuffer)
					data = kerbStoredCredentialNew.Buffer
					for credential in range(kerbStoredCredentialNew.CredentialCount):
						keyDataNew = KERB_KEY_DATA_NEW.from_bytes(data)
						data = data[keyDataNew._len:]
						keyValue = propertyValueBuffer[keyDataNew.KeyOffset:][:keyDataNew.KeyLength]

						if  keyDataNew.KeyType in KERBEROS_TYPE:
							#secret.kerberos_keys.append()
							answer =  (KERBEROS_TYPE[keyDataNew.KeyType],keyValue)
						else:
							answer =  (hex(keyDataNew.KeyType), keyValue)
						
						secret.kerberos_keys.append(answer)

				elif userprop.PropertyName == 'Primary:CLEARTEXT':
					# [MS-SAMR] 3.1.1.8.11.5 Primary:CLEARTEXT Property
					# This credential type is the cleartext password. The value format is the UTF-16 encoded cleartext password.
					try:
						secret.cleartext_pwds.append(userprop.PropertyValue.decode('utf-16le'))
					except UnicodeDecodeError:
						# This could be because we're decoding a machine password. Printing it hex
						secret.cleartext_pwds.append(userprop.PropertyValue.decode('utf-8'))

			return True, None
		except Exception as e:
			return None, e

	async def dump_secrets(self, only_ntlm = False, with_history = True):
		try:
			_, err = await self.get_pek()
			if err is not None:
				raise err

			db_cursor, err = await self.db.open_table('datatable')
			if err is not None:
				raise err
			
			while True:
				await asyncio.sleep(0)
				record, err = await self.db.get_next_row(db_cursor, NTDS_SUPPORTED_SECRET_TYPES)
				if err is not None:
					raise err
				
				if record is None:
					#we parsed all
					return

				if len(record) != 0 and record[NAME_TO_INTERNAL['sAMAccountType']] in ACCOUNT_TYPES and NAME_TO_INTERNAL['objectSid'] in record:
					secret = UserSecrets()
					hash, err = self.__decrypt_hash(secret, record, db_cursor, with_history)
					if err is not None:
						yield None, None
						continue
						#raise err
					if NAME_TO_INTERNAL['supplementalCredentials'] in record:
						_, err = self.__decryptSupplementalInfo(secret, record)
						if err is not None:
							raise err
				
					yield secret, None
					continue
				else:
					#for progress bar
					yield None, None

		except Exception as e:
			yield None, e



from aesedb.security.constants import INTERNAL_TO_NAME, NAME_TO_ATTRTYP
from collections import OrderedDict
import enum
from struct import unpack

from aesedb.structures.header import ESENT_DB_HEADER
from aesedb.structures.page import ESENT_PAGE, PAGEFLAGS, WELLKNOWNPAGENO
from aesedb.structures.branchentry import readbranchentry
from aesedb.structures.leafentry import readleafentry
from aesedb.structures.datadefinitionheader import ESENT_DATA_DEFINITION_HEADER
from aesedb.structures.catalogdatadefentry import ESENT_CATALOG_DATA_DEFINITION_ENTRY, CATALOG_TYPE


TABLE_CURSOR = {
	'TableData' : b'',
	'FatherDataPageNumber': 0,
	'CurrentPageData' : b'',
	'CurrentTag' : 0,
}

# Tagged Data Type Flags
class TAGGED_DATA_TYPE(enum.IntFlag):
	VARIABLE_SIZE = 1
	COMPRESSED    = 2
	STORED        = 4
	MULTI_VALUE   = 8
	WHO_KNOWS     = 10

# Column Types
JET_coltypNil          = 0
JET_coltypBit          = 1
JET_coltypUnsignedByte = 2
JET_coltypShort        = 3
JET_coltypLong         = 4
JET_coltypCurrency     = 5
JET_coltypIEEESingle   = 6
JET_coltypIEEEDouble   = 7
JET_coltypDateTime     = 8
JET_coltypBinary       = 9
JET_coltypText         = 10
JET_coltypLongBinary   = 11
JET_coltypLongText     = 12
JET_coltypSLV          = 13
JET_coltypUnsignedLong = 14
JET_coltypLongLong     = 15
JET_coltypGUID         = 16
JET_coltypUnsignedShort= 17
JET_coltypMax          = 18

ColumnTypeSize = {
    JET_coltypNil          : None,
    JET_coltypBit          : (1,'B'),
    JET_coltypUnsignedByte : (1,'B'),
    JET_coltypShort        : (2,'<h'),
    JET_coltypLong         : (4,'<l'),
    JET_coltypCurrency     : (8,'<Q'),
    JET_coltypIEEESingle   : (4,'<f'),
    JET_coltypIEEEDouble   : (8,'<d'),
    JET_coltypDateTime     : (8,'<Q'),
    JET_coltypBinary       : None,
    JET_coltypText         : None, 
    JET_coltypLongBinary   : None,
    JET_coltypLongText     : None,
    JET_coltypSLV          : None,
    JET_coltypUnsignedLong : (4,'<L'),
    JET_coltypLongLong     : (8,'<Q'),
    JET_coltypGUID         : (16,'16s'),
    JET_coltypUnsignedShort: (2,'<H'),
    JET_coltypMax          : None,
}

# Code pages
CODEPAGE_UNICODE = 1200
CODEPAGE_ASCII   = 20127
CODEPAGE_WESTERN = 1252

StringCodePages = {
    CODEPAGE_UNICODE : 'utf-16le', 
    CODEPAGE_ASCII   : 'ascii',
    CODEPAGE_WESTERN : 'cp1252',
}

class ESENT_DB:
	def __init__(self, dbfile, pageSize = 8192):
		self.__pageSize = pageSize
		self.__DB = dbfile
		self.__DBHeader = None
		self.__totalPages = None
		self.__tables = OrderedDict()
		self.__currentTable = None
	
	async def parse(self):
		try:
			mainHeader, err = await self.get_page(-1)
			if err is not None:
				raise err
			self.__DBHeader = ESENT_DB_HEADER.from_bytes(mainHeader)
			#print(self.__DBHeader)
			self.__pageSize = self.__DBHeader.PageSize
			await self.__DB.seek(0,2)
			self.__totalPages = (self.__DB.tell() // self.__pageSize) -2
			#print("Database Version:0x%x, Revision:0x%x"% (self.__DBHeader.Version, self.__DBHeader.FileFormatRevision))
			#print("Page Size: %d" % self.__pageSize)
			#print("Total Pages in file: %d" % self.__totalPages)
			_, err = await self.parse_catalog(WELLKNOWNPAGENO.CATALOG.value)
			if err is not None:
				raise err

			return True, None
		except Exception as e:
			return None, e
	
	async def get_page(self, pageno, force_parse = False):
		try:
			#print("Trying to fetch page %d (0x%x)" % (pageno, (pageno+1)*self.__pageSize))
			await self.__DB.seek((pageno+1)*self.__pageSize, 0)
			data = await self.__DB.read(self.__pageSize)
			while len(data) < self.__pageSize:
				remaining = self.__pageSize - len(data)
				t = await self.__DB.read(remaining)
				data += t
			# Special case for the first page
			if pageno <= 0 and force_parse is False:
				return data, None
			
			page = ESENT_PAGE(self.__DBHeader)
			_, err = await page.read(data)
			if err is not None:
				raise err
			#print(page)
			return page, None
		except Exception as e:
			return None, e
		
	
	async def parse_catalog(self, pageno):
		try:
			page, err = await self.get_page(pageno)
			if err is not None:
				raise err
			_, err = await self.parse_page(page)
			if err is not None:
				raise err

			for i in range(1, page.record.FirstAvailablePageTag):
				flags, data = page.get_tag(i)
				if page.record.PageFlags & PAGEFLAGS.LEAF == 0:
					# Branch page
					branchEntry, err = await readbranchentry(flags, data)
					if err is not None:
						raise err
						
					_, err = await self.parse_catalog(branchEntry.ChildPageNumber)
					if err is not None:
						raise err
			return True, None
		except Exception as e:
			return None, e
	
	async def parse_page(self, page):
		try:
			for tagno in range(1,page.record.FirstAvailablePageTag):
				flags, data = page.get_tag(tagno)
				if page.record.PageFlags & PAGEFLAGS.LEAF > 0:
					# Leaf page
					if page.record.PageFlags & PAGEFLAGS.SPACE_TREE > 0:
						pass
					elif page.record.PageFlags & PAGEFLAGS.INDEX > 0:
						pass
					elif page.record.PageFlags & PAGEFLAGS.LONG_VALUE > 0:
						pass
					else:
						le, err = await readleafentry(flags, data)
						if err is not None:
							raise err
						self.__addItem(le)
						
					

			return True, None
		except Exception as e:
			return None, e

	def __parseItemName(self,entry):
		dataDefinitionHeader = ESENT_DATA_DEFINITION_HEADER.from_bytes(entry.EntryData)

		if dataDefinitionHeader.LastVariableDataType > 127:
			numEntries =  dataDefinitionHeader.LastVariableDataType - 127
		else:
			numEntries =  dataDefinitionHeader.LastVariableDataType

		itemLen = int.from_bytes(entry.EntryData[dataDefinitionHeader.VariableSizeOffset:][:2], byteorder='little', signed=False)
		itemName = entry.EntryData[dataDefinitionHeader.VariableSizeOffset:][2*numEntries:][:itemLen]
		return itemName

	def __addItem(self, entry):
		dataDefinitionHeader = ESENT_DATA_DEFINITION_HEADER.from_bytes(entry.EntryData)
		catalogEntry = ESENT_CATALOG_DATA_DEFINITION_ENTRY.from_bytes(entry.EntryData[dataDefinitionHeader._len:])
		itemName = self.__parseItemName(entry)

		if catalogEntry.Type == CATALOG_TYPE.TABLE:
			self.__tables[itemName] = OrderedDict()
			self.__tables[itemName].TableEntry = entry
			self.__tables[itemName].Columns    = OrderedDict()
			self.__tables[itemName].Indexes    = OrderedDict()
			self.__tables[itemName].LongValues = OrderedDict()
			self.__currentTable = itemName
		elif catalogEntry.Type == CATALOG_TYPE.COLUMN:
			self.__tables[self.__currentTable].Columns[itemName] = entry
			self.__tables[self.__currentTable].Columns[itemName].Header = dataDefinitionHeader
			self.__tables[self.__currentTable].Columns[itemName].Record = catalogEntry
		elif catalogEntry.Type == CATALOG_TYPE.INDEX:
			self.__tables[self.__currentTable].Indexes[itemName] = entry
		elif catalogEntry.Type == CATALOG_TYPE.LONG_VALUE:
			self.__addLongValue(entry)
		else:
			raise Exception('Unknown type 0x%x' % catalogEntry['Type'])

	def __addLongValue(self, entry):
		dataDefinitionHeader = ESENT_DATA_DEFINITION_HEADER.from_bytes(entry.EntryData)
		lvLen = int.from_bytes(entry.EntryData[dataDefinitionHeader.VariableSizeOffset:][:2], byteorder='little', signed=False)
		lvName = entry.EntryData[dataDefinitionHeader.VariableSizeOffset:][7:][:lvLen]
		self.__tables[self.__currentTable].LongValues[lvName] = entry

	async def open_table(self, tablename):
		try:
			# Returns a cursos for later use

			if isinstance(tablename, bytes) is not True:
				tablename = tablename.encode()

			if tablename in self.__tables:
				entry = self.__tables[tablename].TableEntry
				dataDefinitionHeader = ESENT_DATA_DEFINITION_HEADER.from_bytes(entry.EntryData)
				catalogEntry = ESENT_CATALOG_DATA_DEFINITION_ENTRY.from_bytes(entry.EntryData[dataDefinitionHeader._len:])
				
				# Let's position the cursor at the leaf levels for fast reading
				pageNum = catalogEntry.FatherDataPageNumber
				done = False
				while done is False:
					page, err = await self.get_page(pageNum)
					if err is not None:
						raise err
					
					if page.record.FirstAvailablePageTag <= 1:
						# There are no records
						done = True
					for i in range(1, page.record.FirstAvailablePageTag):
						flags, data = page.get_tag(i)
						if page.record.PageFlags & PAGEFLAGS.LEAF == 0:
							# Branch page, move on to the next page
							branchEntry, err = await readbranchentry(flags, data)
							if err is not None:
								raise err
							pageNum = branchEntry.ChildPageNumber
							break
						else:
							done = True
							break
					
				cursor = TABLE_CURSOR
				cursor['TableData'] = self.__tables[tablename]
				cursor['FatherDataPageNumber'] = catalogEntry.FatherDataPageNumber
				cursor['CurrentPageData'] = page
				cursor['CurrentTag']  = 0
				return cursor, None
			else:
				raise Exception('Table not found!')
		except Exception as e:
			return None, e

	async def __get_next_tag(self, cursor):
		try:
			page = cursor['CurrentPageData']

			if cursor['CurrentTag'] >= page.record.FirstAvailablePageTag:
				# No more data in this page, chau
				return None, None

			flags, data = page.get_tag(cursor['CurrentTag'])
			if page.record.PageFlags & PAGEFLAGS.LEAF > 0:
				# Leaf page
				if page.record.PageFlags & PAGEFLAGS.SPACE_TREE > 0:
					raise Exception('FLAGS_SPACE_TREE > 0')
				elif page.record.PageFlags & PAGEFLAGS.INDEX > 0:
					raise Exception('FLAGS_INDEX > 0')
				elif page.record.PageFlags & PAGEFLAGS.LONG_VALUE > 0:
					raise Exception('FLAGS_LONG_VALUE > 0')
				else:
					# Table Value
					leafEntry, err = await readleafentry(flags, data)
					if err is not None:
						raise err
					return leafEntry, None

			return None, None
		except Exception as e:
			return None, e

	async def get_rowcnt(self, table_name):
		try:
			cursor, err = await self.open_table(table_name)
			if err is not None:
				raise err
			
			ctr = 0
			while True:
				res, err = await self.__next_row_ctr(cursor)
				if err is not None:
					raise err
				if res is None:
					break
				ctr += 1
			
			return ctr, None

		except Exception as e:
			return None, e
	
	async def __next_row_ctr(self, cursor):
		try:
			cursor['CurrentTag'] += 1
			tag, err = await self.__get_next_tag(cursor)
			if err is not None:
				raise err
			
			if tag is None:
				# No more tags in this page, search for the next one on the right
				page = cursor['CurrentPageData']
				if page.record.NextPageNumber == 0:
					# No more pages, chau
					return None, None
				else:
					cursor['CurrentPageData'], err = await self.get_page(page.record.NextPageNumber)
					if err is not None:
						raise err
					cursor['CurrentTag'] = 0
					return await self.__next_row_ctr(cursor)
			return True, None
				

		except Exception as e:
			return None, e
	
	async def get_next_row(self, cursor, filter_col = None):
		try:
			cursor['CurrentTag'] += 1
			#print('currenttag %s ' %cursor['CurrentTag'])
			tag, err = await self.__get_next_tag(cursor)
			if err is not None:
				raise err
			#print(tag)

			if tag is None:
				# No more tags in this page, search for the next one on the right
				page = cursor['CurrentPageData']
				if page.record.NextPageNumber == 0:
					# No more pages, chau
					return None, None
				else:
					cursor['CurrentPageData'], err = await self.get_page(page.record.NextPageNumber)
					if err is not None:
						raise err
					cursor['CurrentTag'] = 0
					return await self.get_next_row(cursor)
			else:
				res, err = await self.__tag_to_record(cursor, tag.EntryData, filter_col)
				if err is not None:
					raise err
				return res, None
		
		except Exception as e:
			return None, e
	
	async def __tag_to_record(self, cursor, tag, filter_col:dict = None):
		try:
			record = OrderedDict()
			taggedItems = OrderedDict()
			taggedItemsParsed = False

			dataDefinitionHeader = ESENT_DATA_DEFINITION_HEADER.from_bytes(tag)
			#dataDefinitionHeader.dump()
			variableDataBytesProcessed = (dataDefinitionHeader.LastVariableDataType - 127) * 2
			prevItemLen = 0
			tagLen = len(tag)
			fixedSizeOffset = dataDefinitionHeader._len
			variableSizeOffset = dataDefinitionHeader.VariableSizeOffset 
	
			columns = cursor['TableData'].Columns

			prev_column_name = None
			prev_column_mandatory = False
			for column in list(columns.keys()):
				#input(column)
				#input(filter_col)
				#if prev_column_name is not None:
				#	if record[prev_column_name] is None and prev_column_mandatory is True:
				#		#if INTERNAL_TO_NAME[prev_column_name].lower() != 'peklist':
				#		#	print(INTERNAL_TO_NAME[prev_column_name])
				#		#	print(record[prev_column_name] is None)
				#		#	print(prev_column_mandatory is True)
				#		#	print('Mandatory field missing! Skipping record!')
				#		#	input()
				#		return OrderedDict(), None

				if filter_col is not None and column not in filter_col:
					#print('skipping')
					continue
				
				if filter_col is not None:
					prev_column_name = column
					prev_column_mandatory = bool(filter_col[column])

				columnRecord = columns[column].Record
				#columnRecord.dump()
				if columnRecord.Identifier <= dataDefinitionHeader.LastFixedSize:
					# Fixed Size column data type, still available data
					record[column] = tag[fixedSizeOffset:][:columnRecord.SpaceUsage]
					fixedSizeOffset += columnRecord.SpaceUsage

				elif 127 < columnRecord.Identifier <= dataDefinitionHeader.LastVariableDataType:
					# Variable data type
					index = columnRecord.Identifier - 127 - 1
					itemLen = int.from_bytes(tag[variableSizeOffset+index*2:][:2], byteorder='little', signed=False)

					if itemLen & 0x8000:
						# Empty item
						itemLen = prevItemLen
						record[column] = None
					else:
						itemValue = tag[variableSizeOffset+variableDataBytesProcessed:][:itemLen-prevItemLen]
						record[column] = itemValue

					#if columnRecord['Identifier'] <= dataDefinitionHeader['LastVariableDataType']:
					variableDataBytesProcessed +=itemLen-prevItemLen

					prevItemLen = itemLen

				elif columnRecord.Identifier > 255:
					# Have we parsed the tagged items already?
					if taggedItemsParsed is False and (variableDataBytesProcessed+variableSizeOffset) < tagLen:
						index = variableDataBytesProcessed+variableSizeOffset
						#hexdump(tag[index:])
						endOfVS = self.__pageSize
						firstOffsetTag = (int.from_bytes(tag[index+2:][:2], byteorder='little', signed=False) & 0x3fff) + variableDataBytesProcessed + variableSizeOffset
						while True:
							taggedIdentifier = int.from_bytes(tag[index:][:2], byteorder='little', signed=False)
							index += 2
							taggedOffset = int.from_bytes(tag[index:][:2], byteorder='little', signed=False) & 0x3fff 
							# As of Windows 7 and later ( version 0x620 revision 0x11) the 
							# tagged data type flags are always present
							if self.__DBHeader.Version == 0x620 and self.__DBHeader.FileFormatRevision >= 17 and self.__DBHeader.PageSize > 8192: 
								flagsPresent = 1
							else:
								flagsPresent = int.from_bytes(tag[index:][:2], byteorder='little', signed=False) & 0x4000
							index += 2
							if taggedOffset < endOfVS:
								endOfVS = taggedOffset
							taggedItems[taggedIdentifier] = (taggedOffset, tagLen, flagsPresent)
							#print "ID: %d, Offset:%d, firstOffset:%d, index:%d, flag: 0x%x" % (taggedIdentifier, taggedOffset,firstOffsetTag,index, flagsPresent)
							if index >= firstOffsetTag:
								# We reached the end of the variable size array
								break
					
						# Calculate length of variable items
						# Ugly.. should be redone
						prevKey = list(taggedItems.keys())[0]
						for i in range(1,len(taggedItems)):
							offset0, length, flags = taggedItems[prevKey]
							offset, _, _ = list(taggedItems.items())[i][1]
							taggedItems[prevKey] = (offset0, offset-offset0, flags)
							#print ("ID: %d, Offset: %d, Len: %d, flags: %d" % (prevKey, offset0, offset-offset0, flags))
							prevKey = list(taggedItems.keys())[i]
						taggedItemsParsed = True
	
					# Tagged data type
					if columnRecord.Identifier in taggedItems:
						offsetItem = variableDataBytesProcessed + variableSizeOffset + taggedItems[columnRecord.Identifier][0] 
						itemSize = taggedItems[columnRecord.Identifier][1]
						# If item have flags, we should skip them
						if taggedItems[columnRecord.Identifier][2] > 0:
							itemFlag = ord(tag[offsetItem:offsetItem+1])
							offsetItem += 1
							itemSize -= 1
						else:
							itemFlag = 0

						#print("ID: %d, itemFlag: 0x%x" %( columnRecord.Identifier, itemFlag))
						if itemFlag & TAGGED_DATA_TYPE.COMPRESSED:
							#print('Unsupported tag column: %s, flag:0x%x' % (column, itemFlag))
							record[column] = None
						elif itemFlag & TAGGED_DATA_TYPE.MULTI_VALUE:
							# ToDo: Parse multi-values properly
							#print('Multivalue detected in column %s, returning raw results' % (column))
							record[column] = (tag[offsetItem:][:itemSize],)
						else:
							record[column] = tag[offsetItem:][:itemSize]

					else:
						record[column] = None
				else:
					record[column] = None

				# If we understand the data type, we unpack it and cast it accordingly
				# otherwise, we just encode it in hex
				if type(record[column]) is tuple:
					# A multi value data, we won't decode it, just leave it this way
					record[column] = record[column][0]
				elif columnRecord.ColumnType == JET_coltypText or columnRecord.ColumnType == JET_coltypLongText: 
					# Let's handle strings
					if record[column] is not None:
						if columnRecord.CodePage not in StringCodePages:
							raise Exception('Unknown codepage 0x%x'% columnRecord['CodePage'])
						stringDecoder = StringCodePages[columnRecord.CodePage]

						try:
							record[column] = record[column].decode(stringDecoder)
						except Exception:
							#print("Exception:", exc_info=True)
							#print('Fixing Record[%r][%d]: %r' % (column, columnRecord['ColumnType'], record[column]))
							record[column] = record[column].decode(stringDecoder, "replace")
							pass
				else:
					unpackData = ColumnTypeSize[columnRecord.ColumnType]
					if record[column] is not None:
						if unpackData is None:
							record[column] = record[column]
						else:
							unpackStr = unpackData[1]
							record[column] = unpack(unpackStr, record[column])[0]

			return record, None

		except Exception as e:
			return None, e




import io


class ESENT_JET_SIGNATURE:
	def __init__(self):
		self.Random = None
		self.CreationTime = None
		self.NetBiosName = b""
	
	@staticmethod
	def from_bytes(data):
		return ESENT_JET_SIGNATURE.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = ESENT_JET_SIGNATURE()
		res.Random = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.CreationTime = int.from_bytes(buff.read(8), byteorder='little', signed=False)
		res.NetBiosName = buff.read(16)
		return res
	


from abc import ABC, abstractmethod
import enum

class cipherMODE(enum.Enum):
	ECB = enum.auto()
	CBC = enum.auto()
	CTR = enum.auto()
	CCM = enum.auto()
	CFB = enum.auto()

class symmetricBASE():
	def __init__(self):
		self._cipher = None
		self.setup_cipher()

	@abstractmethod
	def setup_cipher(self):
		#create the hash object here
		pass

	@abstractmethod
	def encrypt(self, data):
		pass

	@abstractmethod
	def decrypt(self):
		pass

	@abstractmethod
	def update(self, data):
		pass

	@abstractmethod
	def digest(self):
		pass

class hashBASE():
	def __init__(self, data):
		self._hash = None
		self.setup_hash()

		if data is not None:
			self._hash.update(data)

	@abstractmethod
	def setup_hash(self):
		#create the hash object here
		pass

	@abstractmethod
	def update(self, data):
		pass

	@abstractmethod
	def digest(self):
		pass

	@abstractmethod
	def hexdigest(self):
		pass

class hmacBASE():
	def __init__(self, key):
		self._key = key
		self._hash = None
		self.setup_hash()

	@abstractmethod
	def setup_hash(self):
		#create the hash object here
		pass

	@abstractmethod
	def update(self, data):
		pass

	@abstractmethod
	def digest(self):
		pass

	@abstractmethod
	def hexdigest(self):
		pass


"""
The idea here is to offer compatibility with 3rd party libraries by extending wrappers for ech encryption mode
This is needed because the pure python implementation for encryption and hashing algorithms are quite slow

currently it's not the perfect wrapper, needs to be extended
"""

import io
from aesedb.security.crypto.BASE import symmetricBASE, cipherMODE
from aesedb.security.crypto.pure.AES import AESModeOfOperationECB, AESModeOfOperationCBC, AESModeOfOperationCTR, AESModeOfOperationCFB
from aesedb.security.crypto.pure.AES.blockfeeder import decrypt_stream, encrypt_stream, PADDING_NONE
try:
	from Crypto.Cipher import AES as _pyCryptoAES
except:
	pass

try:
	from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
	from cryptography.hazmat.primitives import padding
	from cryptography.hazmat.backends import default_backend
except:
	pass

try:
	from Cryptodome.Cipher import AES as _pyCryptodomeAES
except:
	pass

class pureAES(symmetricBASE):
	def __init__(self, key, mode = cipherMODE.ECB, IV = None, nonce = None, pad = None, padMode = None, segment_size = 8):
		self.key = key
		self.mode = mode
		self.IV = IV
		self.pad = pad
		self.padMode = padMode
		self.nonce = nonce
		self.segment_size = segment_size

		symmetricBASE.__init__(self)

	def setup_cipher(self):
		if self.mode == cipherMODE.ECB:
			self._cipher = AESModeOfOperationECB(self.key)
		elif self.mode == cipherMODE.CBC:
			self._cipher = AESModeOfOperationCBC(self.key, iv = self.IV)
		elif self.mode == cipherMODE.CTR:
			self._cipher = AESModeOfOperationCTR(self.key, counter = self.IV)
		elif self.mode == cipherMODE.CFB:
			self._cipher = AESModeOfOperationCFB(self.key, iv = self.IV, segment_size = self.segment_size)
		else:
			raise Exception('Unknown cipher mode!')

	def encrypt(self, data):
		return self._cipher.encrypt(data)
	
	def decrypt(self, data):
		in_buff = io.BytesIO(data)
		out_buff = io.BytesIO()
		decrypt_stream(self._cipher, in_buff, out_buff, padding = PADDING_NONE)
		out_buff.seek(0)
		return out_buff.read()
	
	def update(self, data):
		if self.mode != cipherMODE.CCM:
			raise Exception('Not applicable!')

		return self._cipher.update(data)

	def digest(self):
		if self.mode != cipherMODE.CCM:
			raise Exception('Not applicable!')

		return self._cipher.digest()



class pyCryptoAES(symmetricBASE):
	def __init__(self, key, mode = cipherMODE.ECB, IV = None, pad = None, padMode = None):
		self.key = key
		self.mode = mode
		self.IV = IV
		self.pad = pad
		self.padMode = padMode

		symmetricBASE.__init__(self)
		
	def setup_cipher(self):
		if self.mode == cipherMODE.ECB:
			self._cipher = _pyCryptoAES.new(self.key, _pyCryptoAES.MODE_ECB)
		
		elif self.mode == cipherMODE.CBC:
			self._cipher = _pyCryptoAES.new(self.key, _pyCryptoAES.MODE_CBC, self.IV)
		elif self.mode == cipherMODE.CTR:
			self._cipher = _pyCryptoAES.new(self.key, _pyCryptoAES.MODE_CTR, self.IV)
		else:
			raise Exception('Unknown cipher mode!')
		
	def encrypt(self, data):
		return self._cipher.encrypt(data)
	def decrypt(self, data):
		return self._cipher.decrypt(data)

class cryptographyAES(symmetricBASE):
	def __init__(self, key, mode = cipherMODE.ECB, IV = None, pad = None, padMode = None):
		self.IV = IV
		#the python cryptography module sets the IV in the operational mode!!!
		if mode == cipherMODE.ECB:
			self.IV = modes.ECB()
		elif mode == cipherMODE.CBC:
			self.IV = modes.CBC(IV)
		elif mode == cipherMODE.CBC:
			self.IV = modes.CTR(IV)
		else:
			raise Exception('Unknown cipher mode!')
		
		self.key = key
		
		""" TODO padding
		if self.padMode is not None:
		"""
		
		self.encryptor = None
		self.decryptor = None
		symmetricBASE.__init__(self)

	def setup_cipher(self):
		algorithm = algorithms.AES(self.key)
		self._cipher = Cipher(algorithm, mode=self.IV, backend=default_backend())
		self.encryptor = self._cipher.encryptor()
		self.decryptor = self._cipher.decryptor()

	def encrypt(self, data):
		return self.encryptor.update(data)
	

	def decrypt(self, data):
		return self.decryptor.update(data)


class pyCryptodomeAES(symmetricBASE):
	def __init__(self, key, mode = cipherMODE.ECB, IV = None):
		self.key = key
		self.mode = mode
		self.IV = IV
		symmetricBASE.__init__(self)

	def setup_cipher(self):
		if self.mode == cipherMODE.ECB:
			self._cipher = _pyCryptodomeAES.new(self.key, _pyCryptodomeAES.MODE_ECB) #_pyCryptodomeDES.new(self.key, _pyCryptodomeDES.MODE_ECB)
		elif self.mode == cipherMODE.CBC:
			self._cipher = _pyCryptodomeAES.new(self.key, _pyCryptodomeAES.MODE_CBC, iv=self.IV)
		elif self.mode == cipherMODE.CTR:
			self._cipher = _pyCryptodomeAES.new(self.key, _pyCryptodomeAES.MODE_CTR, iv=self.IV)
		else:
			raise Exception('Unknown cipher mode!')
		
	def encrypt(self, data):
		return self._cipher.encrypt(data)
	def decrypt(self, data):
		return self._cipher.decrypt(data)


import io

class CRYPTED_HASH:
	def __init__(self):
		self.Header = None
		self.KeyMaterial = None
		self.EncryptedHash = None

	@staticmethod
	def from_bytes(data):
		return CRYPTED_HASH.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = CRYPTED_HASH()
		res.Header = buff.read(8)
		res.KeyMaterial = buff.read(16)
		res.EncryptedHash = buff.read(16)
		return res
	
class CRYPTED_HASHW16:
	def __init__(self):
		self.Header = None
		self.KeyMaterial = None
		self.Unknown = None
		self.EncryptedHash = None

	@staticmethod
	def from_bytes(data):
		return CRYPTED_HASHW16.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = CRYPTED_HASHW16()
		res.Header = buff.read(8)
		res.KeyMaterial= buff.read(16)
		res.Unknown = buff.read(4)
		res.EncryptedHash = buff.read()
		return res

class CRYPTED_HISTORY:
	def __init__(self):
		self._len = 24
		self.Header = None
		self.KeyMaterial = None
		self.EncryptedHash = None

	@staticmethod
	def from_bytes(data):
		return CRYPTED_HISTORY.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = CRYPTED_HISTORY()
		res.Header = buff.read(8)
		res.KeyMaterial = buff.read(16)
		res.EncryptedHash = buff.read()
		res._len += len(res.EncryptedHash)
		return res

class CRYPTED_BLOB:
	def __init__(self):
		self.Header = None
		self.KeyMaterial = None
		self.EncryptedHash = None

	@staticmethod
	def from_bytes(data):
		return CRYPTED_BLOB.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = CRYPTED_BLOB()
		res.Header = buff.read(8)
		res.KeyMaterial= buff.read(16)
		res.EncryptedHash = buff.read()
		return res


import io

class USER_PROPERTIES:
	def __init__(self):
		self.Reserved1 = None
		self.Length  = None
		self.Reserved2  = None
		self.Reserved3 = None
		self.Reserved4 = None
		self.PropertySignature = None
		self.PropertyCount = None
		self.UserProperties = []
	
	@staticmethod
	def from_bytes(data):
		return USER_PROPERTIES.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = USER_PROPERTIES()
		res.Reserved1 = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.Length  = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.Reserved2  = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.Reserved3 = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.Reserved4 = buff.read(96)
		res.PropertySignature = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.PropertyCount = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		for _ in range(res.PropertyCount):
			prop = USER_PROPERTY.from_buffer(buff)
			res.UserProperties.append(prop)
		return res
	
	def __str__(self):
		t = ''
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for e in self.__dict__[k]:
					t += '%s : %s %s' % (k, self.__dict__[k], str(e))
			else:
				t += '%s : %s' % (k, self.__dict__[k])
		return t


class USER_PROPERTY:
	def __init__(self):
		self.NameLength = None
		self.ValueLength = None
		self.Reserved = None
		self.PropertyName = None
		self.PropertyValue = None
	
	@staticmethod
	def from_bytes(data):
		return USER_PROPERTY.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = USER_PROPERTY()
		res.NameLength = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.ValueLength = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.Reserved = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.PropertyName = buff.read(res.NameLength).decode('utf-16le')
		res.PropertyValue = buff.read(res.ValueLength)
		return res

	def __str__(self):
		t = ''
		for k in self.__dict__:
			t += '%s : %s' % (k, self.__dict__[k])
		return t


import io
import enum

class ESENT_TAGS(enum.IntFlag):
	UNKNOWN = 0x1
	DEFUNCT = 0x2
	COMMON  = 0x4


class ESENT_BRANCH_ENTRY_COMMON:
	def __init__(self):
		self.CommonPageKeySize = None
		self.LocalPageKeySize = None
		self.LocalPageKey = None
		self.ChildPageNumber = None
	
	@staticmethod
	def from_bytes(data):
		return ESENT_BRANCH_ENTRY_COMMON.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = ESENT_BRANCH_ENTRY_COMMON()
		res.CommonPageKeySize = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.LocalPageKeySize = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.LocalPageKey = buff.read(res.LocalPageKeySize)
		res.ChildPageNumber = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		return res

class ESENT_BRANCH_ENTRY:
	def __init__(self):
		self.LocalPageKeySize = None
		self.LocalPageKey = None
		self.ChildPageNumber = None
	
	@staticmethod
	def from_bytes(data):
		return ESENT_BRANCH_ENTRY.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = ESENT_BRANCH_ENTRY()
		res.LocalPageKeySize = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.LocalPageKey = buff.read(res.LocalPageKeySize)
		res.ChildPageNumber = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		return res

async def readbranchentry(flags, data):
	try:
		if ESENT_TAGS.COMMON in ESENT_TAGS(flags):
			return ESENT_BRANCH_ENTRY_COMMON.from_bytes(data), None
		return ESENT_BRANCH_ENTRY.from_bytes(data), None
	except Exception as e:
		return None, e